"""
Just useful objects to make your coding nicer every day
"""

from .algos import merge_dicts
from .concurrent import Monitor, RMonitor
from .decorators import precondition, short_none, has_keys, \
    wraps, chain_functions, postcondition, queue_get, auto_adapt_to_methods, \
    attach_arguments, for_argument
from .deleters import ListDeleter, DictDeleter
from .fun_static import static_var
from .iterators import hint_with_length, SelfClosingGenerator, exhaust, chain
from .metaclasses import metaclass_maker, wrap_with, dont_wrap, wrap_property, DocsFromParent, \
    CopyDocsFrom
from .misc import update_if_not_none, update_key_if_none, update_attr_if_none, queue_iterator, \
    update_key_if_not_none, source_to_function, update_key_if_true, \
    get_arguments, call_with_arguments, chain_callables, Closeable, contains, \
    enum_value
from .overloading import overload, class_or_instancemethod
from .recast_exceptions import rethrow_as, silence_excs, catch_exception, log_exceptions, \
    raises_exception
from .expect_exception import expect_exception

__all__ = [
    'Closeable', 'contains', 'enum_value',
    'expect_exception',
    'overload', 'class_or_instancemethod',
    'update_if_not_none', 'DocsFromParent', 'update_key_if_none', 'queue_iterator',
    'update_attr_if_none', 'update_key_if_not_none', 'source_to_function',
    'update_key_if_true',
    'hint_with_length', 'SelfClosingGenerator', 'exhaust', 'chain',
    'Monitor', 'RMonitor', 'merge_dicts', 'chain_callables',
    'short_none', 'has_keys', 'CopyDocsFrom',
    'precondition', 'postcondition', 'queue_get',
    'rethrow_as', 'silence_excs', 'raises_exception',
    'static_var', 'metaclass_maker',
    'catch_exception', 'wraps', 'wrap_with', 'dont_wrap', 'wrap_property', 'log_exceptions',
    'chain_functions', 'get_arguments', 'call_with_arguments',
    'ListDeleter', 'DictDeleter', 'for_argument', 'attach_arguments', 'auto_adapt_to_methods',
]



import sys
import time
import typing as tp

from .exception_handlers import BaseExceptionHandler, ALWAYS_FIRST, ExceptionHandlerCallable
from ..os.misc import suicide


class MemoryErrorExceptionHandler(BaseExceptionHandler):
    """
    A handler that terminates the entire process (or process group) is a MemoryError is seen.

    `custom_hook` is an exception callable to implement you own behavior. If it returns True,
    then MemoryErrorExceptionHandler won't kill anyone. You can also provide a CallableGroup
    with gather=True - if any of callables returns True, the process won't be killed.

    :param kill_pg: whether to kill entire process group, if applicable
    """
    __slots__ = ('priority', '_free_on_memory_error', 'custom_hook', 'kill_pg', 'installed')

    def __init__(self,
                 custom_hook: ExceptionHandlerCallable = lambda type_, value, traceback: False,
                 kill_pg: bool = False):
        super().__init__(ALWAYS_FIRST)
        # so that we have some spare space in case a MemoryError is thrown
        self._free_on_memory_error = {'a': bytearray(1024 * 4)}  # type: tp.Dict[str, bytearray]
        self.custom_hook = custom_hook  # type: ExceptionHandlerCallable
        self.kill_pg = kill_pg  # type: bool
        self.installed = False  # type: bool

    def install(self):
        if self.installed:
            raise RuntimeError('already installed')

        from .global_eh import GlobalExcepthook
        GlobalExcepthook().add_hook(self)

    def handle_exception(self, type_, value, traceback) -> tp.Optional[bool]:
        if not issubclass(type_, MemoryError):
            return

        try:
            del self._free_on_memory_error['a']
        except KeyError:
            pass

        # noinspection PyBroadException
        try:
            val = self.custom_hook(type_, value, traceback)
            if isinstance(val, tp.Sequence):
                val = any(val)

            if val:
                return True
        except Exception:
            pass

        try:
            sys.stderr.write(
                'satella.exception_handling: MemoryError seen, killing\n')
            sys.stderr.flush()
        except (IOError, OSError):
            pass

        suicide(self.kill_pg)

        time.sleep(5)
        return True
